@@ -570,17 +570,17 @@ class ContactService(BaseRdbmsService):
             return
 
         # Check whether this contact would been added to another user account
-        # and wether it would have been verified.
+        # and whether it would have been verified.
         (is_available, is_verified) = self.is_contact_available(contact, connection=connection)
         if not is_available and is_verified:
             raise self.ContactAlreadyUsedException("Another user has already verified this contact information")
 
         # Check whether this contact information can been stated verified.
         if contact.is_verified and not has_been_verified and account.account_type not in [
-                AccountType.botnet,
-                AccountType.sns,
-                AccountType.test
-            ]:
+            AccountType.botnet,
+            AccountType.sns,
+            AccountType.test
+        ]:
             raise self.InvalidArgumentException("A contact cannot be stated as verified without passing a challenge process")
 
         # Add this contact information to the specified user account, and
@@ -627,19 +627,19 @@ class ContactService(BaseRdbmsService):
             # Add this contact information to the specified user account.
             connection.execute(
                 """
-                INSERT INTO
-                    account_contact(
-                        account_id,
-                        name,
-                        value,
-                        is_primary,
-                        is_verified)
-                  VALUES 
-                    (%(account_id)s,
-                     %(name)s,
-                     %(value)s,
-                     %(is_primary)s,
-                     %(is_verified)s)
+                INSERT INTO account_contact(
+                    account_id,
+                    name,
+                    value,
+                    is_primary,
+                    is_verified)
+                  VALUES (
+                    %(account_id)s,
+                    %(name)s,
+                    %(value)s,
+                    %(is_primary)s,
+                    %(is_verified)s
+                  )
                 """,
                 {
                     'account_id': account.account_id,
@@ -647,7 +647,8 @@ class ContactService(BaseRdbmsService):
                     'name': contact.name,
                     'value': contact.value,
                     'is_verified': contact.is_verified or False
-                })
+                }
+            )
 
             # If this contact information needs to be verified, generate a
             # verification request that will be sent to the user.
@@ -835,6 +836,57 @@ class ContactService(BaseRdbmsService):
 
                 return contact, session
 
+    def find_account(self, contact, connection=None):
+        """
+        Return the account that matches a contact.
+
+
+        :param contact: An object `Contact`.
+
+        :param connection: An object `RdbmsConnection`.
+
+
+        :return: `None` if no account matches the specified contact, or an
+            object containing the following attributes:
+
+            * `account_id` (required): Identification of the account.
+
+            * `is_primary` (required): Indicate whether this contact information is
+              the primary one for the given property.
+
+            * `is_verified` (required): Indicate whether this contact information
+              has been verified.
+
+            * `object_status` (required): Current status of this contact information.
+        """
+        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
+            cursor = connection.execute(
+                """
+                SELECT
+                    account_id,
+                    is_primary,
+                    is_verified,
+                    object_status
+                  FROM
+                    account_contact
+                  WHERE
+                    name = %(property_name)s
+                    AND value = %(property_value)s
+                """,
+                {
+                    'property_name': contact.name,
+                    'property_value': contact.value,
+                }
+            )
+
+            row = cursor.fetch_one()
+            account = row and row.get_object({
+                'account_id': cast.string_to_uuid,
+                'object_status': ObjectStatus
+            })
+
+            return account
+
     def get_contact_verification_request(self, contact):
         """
         Return the request send to a user to verify his contact information.
@@ -885,20 +937,24 @@ class ContactService(BaseRdbmsService):
                 'request_id': cast.string_to_uuid
             })
 
-    def get_contacts(self, account_id,
-            connection=None):
+    def get_contacts(
+            self,
+            account_id,
+            connection=None,
+            property_name=None):
         """
-        Fetch the list of contact information that have been added to the
-        specified user account.
+        Return the list of contact information of an account
 
 
-        :param account_id: identification of a user account.
+        :param account_id: Identification of a user account.
 
-        :param connection: a `RdbmsConnection` instance to be used
-            supporting the Python clause `with ...:`.
+        :param connection: An object `RdbmsConnection`.
+
+        :param property_name: An item of the enumeration `ContactName` to
+            filter the list of contact information of the user.
 
 
-        :return: a list of instances `Contact`.
+        :return: A list of `Contact` objects.
         """
         with self.acquire_rdbms_connection(connection=connection) as connection:
             cursor = connection.execute(
@@ -913,13 +969,20 @@ class ContactService(BaseRdbmsService):
                   WHERE
                     account_id = %(account_id)s
                     AND object_status = %(OBJECT_STATUS_ENABLED)s
+                    AND (%(property_name)s IS NULL OR name = %(property_name)s) 
                 """,
                 {
                     'OBJECT_STATUS_ENABLED': ObjectStatus.enabled,
-                    'account_id': account_id
+                    'account_id': account_id,
+                    'property_name': property_name,
                 })
 
-            return [Contact.from_object(row.get_object()) for row in cursor.fetch_all()]
+            contacts = [
+                Contact.from_object(row.get_object())
+                for row in cursor.fetch_all()
+            ]
+
+            return contacts
 
     def is_contact_verification_request(
             self,
@@ -961,49 +1024,73 @@ class ContactService(BaseRdbmsService):
 
             return cursor.get_row_count() > 0
 
-    def set_primary_contact(self, account_id, contact,
+    def set_primary_contact(
+            self,
+            account_id,
+            contact,
+            add_if_not_exists=False,
             connection=None):
         """
-        Set the specified contact information as the primary for the given
-        user account.  This contact information MUST have been verified.  All
-        the other contact information of the same type are set as alternates.
+        Set the primary contact information of a users
+
+        This new or the existing contact information MUST have been verified.
+
+        The function sets all the contact information of the same type as
+        alternates.
 
 
-        :param account_id: identification of the account of a user.
+        :param account_id: The identification of the account of a user.
 
-        :param contact: an instance `Contact` of the contact information of
-            the user to set as primary.
+        :param contact: An object `Contact` of the contact information of
+            the user to set as primary.  The attribute `is_verified` of this
+            object MUST be `True`.
 
-        :param connection: a `RdbmsConnection` instance to be used
-            supporting the Python clause `with ...:`.
+        :param add_if_not_exists: Indicate whether to add this contact
+            information if it doesn't not exist.
+
+        :param connection: An object `RdbmsConnection` with auto commit.
 
 
-        :raise InvalidOperationException: if this contact information has not
+        :raise InvalidOperationException: If the contact information has not
             been verified yet.
 
-        :raise UndefinedObjectException: if this contact information has not
+        :raise UndefinedObjectException: If this contact information has not
             been associated to this user account.
         """
+        if not contact.is_verified:
+            raise self.InvalidArgumentException("The contact information MUST have been verified")
 
-        # Retrieve the contact information of this user account.
-        contacts = self.get_contacts(account_id, connection=connection)
+        # Check whether this contact information is already associated to this
+        # user.  If not, possibly add this contact information to the user; if
+        # already associated to another user, this would raise an exception.
+        contacts = self.get_contacts(
+            account_id,
+            connection=connection,
+            property_name=contact.name)
 
-        # Check whether this contact information is defined for this user
-        # account.
-        try:
-            _contact = [ _contact for _contact in contacts if _contact == contact ][0]
-        except IndexError:
-            raise self.UndefinedObjectException('This contact information is not associated to this user account')
+        existing_contact = [_ for _ in contacts if _.value == contact.value]
+        if len(existing_contact) == 0:
+            if not add_if_not_exists:
+                raise self.UndefinedObjectException('This contact information is not associated to this user account')
+
+            self.add_contact(
+                account_id,
+                contact,
+                connection=connection,
+                has_been_verified=contact.is_verified)
+        else:
+            existing_contact = existing_contact[0]
 
         # Ignore this request if this contact information is already set as the
         # primary contact for this user.
-        if _contact.is_primary:
+        if existing_contact.value == contact.value and existing_contact.is_primary:
             return
 
-        # Check whether this contact information has been verified, as this is
-        # required to set a contact as the primary.
-        if not _contact.is_verified:
-            raise self.InvalidOperationException('A contact information that is not verified cannot be set as the primary')
+        # Check whether the contact information already associated to the user
+        # has been verified (requirement for primary contact).
+        if not existing_contact.is_verified:
+            raise self.InvalidOperationException(
+                "A contact information that was not verified cannot be set as primary")
 
         # Update all the contact information of this type for this user account,
         # setting this particular contact information as the primary, and the
@@ -1014,100 +1101,15 @@ class ContactService(BaseRdbmsService):
                 UPDATE 
                     account_contact
                   SET
-                    is_primary = (value = %(value)s),
+                    is_primary = (value = %(property_value)s),
                     update_time = current_timestamp
                   WHERE
-                     account_id = %(account_id)s
-                    AND name = %(name)s
-                    AND is_primary <> (value = %(value)s)
-                    AND object_status NOT IN (%(OBJECT_STATUS_DELETED)s, %(OBJECT_STATUS_DISABLED)s)
+                    account_id = %(account_id)s
+                    AND name = %(property_name)s
+                    AND is_primary <> (value = %(property_value)s)
                 """,
                 {
-                    'OBJECT_STATUS_DELETED': ObjectStatus.deleted,
-                    'OBJECT_STATUS_DISABLED': ObjectStatus.disabled,
                     'account_id': account_id,
-                    'name': contact.name,
-                    'value': contact.value
+                    'property_name': contact.name,
+                    'property_value': contact.value,
                 })
-
-
-    # def update_contacts(self, app_id, account_id, contacts):
-    #     """
-    #
-    #
-    #     :param app_id: identification of the client application such as a Web,
-    #         a desktop, or a mobile application, that accesses the service.
-    #
-    #     :param account_id: identification of the account of the user whose
-    #         contact information is updated.
-    #
-    #     :param contacts: a list of contact information of the user's account.
-    #         A contact information corresponds to the possible following tuples::
-    #
-    #                `(name, value)`
-    #
-    #         or::
-    #
-    #                `(name, value, is_primary)`
-    #
-    #         or::
-    #
-    #                `(name, value, is_primary, is_verified)`
-    #
-    #         where:
-    #
-    #         * `name`: an item of the enumeration `ContactName`.
-    #
-    #         * `value`: value of the property representing by a string.
-    #
-    #         * `is_primary`: indicate whether this contact information is the
-    #           primary.  By default, the first contact information of a given
-    #           type is the primary.
-    #
-    #         * `is_verified`: indicate whether this contact information has
-    #           been verified, more likely grabbed from a trusted Social
-    #           Networking Service (SNS).  This parameter MUST NOT be surfaced
-    #           by any public API.
-    #
-    #         A contact information is optional if a username is provided, or if
-    #         the account is a mapping to a 3rd party Social Network Service
-    #         (SNS) account.
-    #     :return:
-    #     """
-    #     account = self.get_account(account_id, check_status=True, include_contacts=True)
-    #
-    #     contacts = contact.format_contacts(contacts)
-    #
-    #     with self.acquire_rdbms_connection(auto_commit=True) as connection:
-    #         connection.execute("""
-    #             DELETE FROM account_contact
-    #               WHERE account_id = %(account_id)s
-    #                 AND property_name IN (%[property_names]s)""",
-    #             { 'account_id': account_id,
-    #               'property_names': set([ name for (name, _, _, _) in contacts ]) })
-    #
-    #         for (name, value, is_primary, is_verified) in contacts:
-    #             (is_available, is_existing_contact_verified) = self.is_contact_available(app_id, (name, value),
-    #                     account_id=account_id)
-    #
-    #             if not is_available:  # and is_existing_contact_verified:
-    #                 raise self.ContactAlreadyUsedException('This contact information is already associated with a user account',
-    #                         payload={ 'contact': (name, value) })
-    #
-    #         if is_verified and account.account_type not in [
-    #                 AccountType.botnet,
-    #                 AccountType.sns,
-    #                 AccountType.test]:
-    #             raise self.InvalidArgumentException('Contact of account other than botnet, SNS, and test cannot be stated as verified')
-    #
-    #         connection.execute("""
-    #             INSERT INTO account_contact(
-    #                     account_id,
-    #                     property_name,
-    #                     property_value,
-    #                     is_primary,
-    #                     is_verified,
-    #                     app_id)
-    #               VALUES %[values]s""",
-    #             {'values': [ (account_id, name, value, is_primary, is_verified, app_id)
-    #                     for (name, value, is_primary, is_verified,) in contacts ] })
